# -*- coding: utf-8 -*-
"""
configuration services module.
"""

from pyrin.application.services import get_component
from pyrin.configuration import ConfigurationPackage


def load_configuration(name, **options):
    """
    loads the given configuration if relevant file is
    available in settings path.

    :param str name: configuration name.

    :keyword bool silent: specifies that if a related configuration file
                          for the given name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword bool ignore_on_existed: specifies that it should not raise an
                                     error if a config store with given name
                                     has been already loaded.
                                     defaults to False if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationStoreExistedError: configuration store existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).load_configuration(name,
                                                                                 **options)


def load_configurations(*names, **options):
    """
    loads the given configurations if relevant files is
    available in settings path.

    :param str names: configuration names as arguments.

    :keyword bool silent: specifies that if a related configuration file
                          for any of the given names not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword bool ignore_on_existed: specifies that it should not raise an
                                     error if a config store with given name
                                     has been already loaded.
                                     defaults to False if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationStoreExistedError: configuration store existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).load_configurations(*names,
                                                                                  **options)


def reload(store_name, **options):
    """
    reloads the configuration store from it's relevant file.

    :param str store_name: config store name to be reloaded.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).reload(store_name, **options)


def get_file_path(store_name, **options):
    """
    gets the configuration file path for given config store.

    :param str store_name: config store name to get it's file path.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_file_path(store_name,
                                                                            **options)


def get(store_name, section, key, **options):
    """
    gets the value of specified key from provided section of given config store.

    :param str store_name: config store name.
    :param str section: config section name.
    :param str key: config key to get it's value.

    :keyword object default_value: default value if key not present in config section.
                                   if not provided, error will be raised.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get(store_name, section,
                                                                  key, **options)


def get_active(store_name, key, **options):
    """
    gets the value of given key from active section of given config store.
    if this store does not have an active section, it raises an error.

    :param str store_name: config store name.
    :param str key: config key to get it's value.

    :keyword object default_value: default value if key not present in config section.
                                   if not provided, error will be raised.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active(store_name,
                                                                         key, **options)


def get_section_names(store_name, **options):
    """
    gets all available section names of given config store.

    :param str store_name: config store name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :rtype: list[str]
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section_names(store_name,
                                                                                **options)


def get_section(store_name, section, **options):
    """
    gets all key/values stored in given section of specified config store.

    :param str store_name: config store name.
    :param str section: section name.

    :keyword callable converter: a callable to use as case converter for keys.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section(store_name, section,
                                                                          **options)


def get_section_keys(store_name, section, **options):
    """
    gets all available keys in given section of specified config store.

    :param str store_name: config store name.
    :param str section: section name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :rtype: list[str]
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section_keys(store_name,
                                                                               section,
                                                                               **options)


def get_all(store_name, **options):
    """
    gets all available key/values from different sections of
    given config store in a flat dict, eliminating the sections.
    note that if there are same key names in different
    sections, it raises an error to prevent overwriting values.
    also note that if the config store contains `active` section,
    then the result of `get_active_section` method would be returned.

    :param str store_name: config store name.

    :keyword callable converter: a callable to use as case converter for keys.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.
    :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_all(store_name, **options)


def get_active_section(store_name, **options):
    """
    gets the active section available in given config store.
    this method gets the section that it's name is under [active]
    section, for example:

    [active]
    selected: production

    [production]
    id: 123
    name: prod

    [development]
    id: 233
    name: dev

    this will return all key/values available under [production].
    if the config store has not an [active] section, this method
    raises an error.

    :param str store_name: config store name.

    :keyword callable converter: a callable to use as case converter.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active_section(store_name,
                                                                                 **options)


def get_active_section_name(store_name):
    """
    gets the active section name of given config store if available.
    if the store does not have an active section, it raises an error.

    :param str store_name: config store name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.

    :rtype: str
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active_section_name(store_name)



# -*- coding: utf-8 -*-
"""
datetime services module.
"""

from pyrin.application.services import get_component
from pyrin.globalization.datetime import DateTimePackage


def now(timezone=None):
    """
    gets the current datetime based on given timezone name.

    :param str timezone: timezone name to get current datetime based on it.
                         if not provided, defaults to application
                         current timezone.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).now(timezone)


def get_current_timezone():
    """
    gets the application current timezone.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timezone()


def get_current_timezone_name():
    """
    gets the application current timezone name.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timezone_name()


def normalize(value):
    """
    normalizes input datetime value using application current timezone.
    input value should already have a timezone info.

    :param datetime value: value to get normalized.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).normalize(value)


def localize(value):
    """
    localizes input datetime with application current timezone.
    input value should not have a timezone info.

    :param datetime value: value to be localized.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).localize(value)


def to_datetime_string(value):
    """
    gets the datetime string representation of input value.
    if the value has no timezone info, it adds the application
    current timezone info into it.
    example: `2015-12-24T22:40:15+01:00`

    :param datetime value: input object to be converted.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime_string(value)


def to_date_string(value):
    """
    gets the date string representation of input value.
    example: `2015-12-24`

    :param Union[datetime, date] value: input object to be converted.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date_string(value)


def to_time_string(value):
    """
    gets the time string representation of input value.
    if the value has no timezone info, it adds the application
    current timezone info into it.
    example: `23:40:15`

    :param Union[datetime, time] value: input object to be converted.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time_string(value)


def to_datetime(value):
    """
    converts the input value to it's equivalent python datetime.

    :param str value: string representation of datetime to be converted.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime(value)


def to_date(value):
    """
    converts the input value to it's equivalent python date.

    :param str value: string representation of date to be converted.

    :rtype: date
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date(value)


def to_time(value):
    """
    converts the input value to it's equivalent python time.

    :param str value: string representation of time to be converted.

    :rtype: time
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time(value)


def timezone_exists(timezone_name):
    """
    gets a value indicating that a timezone with the given name exists.

    :param str timezone_name: timezone name to check for existence.

    :rtype: bool
    """

    return get_component(DateTimePackage.COMPONENT_NAME).timezone_exists(timezone_name)


def get_current_timestamp(date_sep='-', main_sep=' ',
                          time_sep=':', timezone=None):
    """
    gets the current timestamp with specified separators based on given timezone.

    :param Union[str, None] date_sep: a separator to put between date elements.
    :param Union[str, None] main_sep: a separator to put between date and time part.
    :param Union[str, None] time_sep: a separator to put between time elements.

    :param Union[str, None] timezone: timezone name to get current timestamp
                                      based on it. if not provided, defaults
                                      to application current timezone.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timestamp(date_sep,
                                                                               main_sep,
                                                                               time_sep,
                                                                               timezone)



# -*- coding: utf-8 -*-
"""
utils environment module.
"""

import platform


def get_os_name():
    """
    gets current os name.

    :rtype: str
    """

    return platform.uname()[0].lower()


def is_windows():
    """
    gets a value indicating that underlying operating system is windows based.

    :rtype: bool
    """

    return 'win' in get_os_name()



# -*- coding: utf-8 -*-
"""
core context module.
"""

from threading import Lock

from pyrin.core.exceptions import CoreAttributeError, ContextAttributeError
from pyrin.utils.singleton import MultiSingletonMeta


class DTO(dict):
    """
    context class for storing objects in every layer.
    it's actually a dictionary with the capability to treat keys as instance attributes.
    this class's objects are immutable and could be used as a dict key if needed.
    """

    def __getattr__(self, name):
        if name in self:
            return self.get(name)

        raise CoreAttributeError('Property [{name}] not found.'.format(name=name))

    def __getitem__(self, item):
        if item in self:
            return self.get(item)

        raise CoreAttributeError('Property [{name}] not found.'.format(name=item))

    def __setattr__(self, name, value):
        self[name] = value

    def __hash__(self):
        signature = None
        try:
            # try to sort by keys if all keys are same type.
            signature = sorted(self.items())
        except TypeError:
            # fallback to unsorted keys when key types are incomparable.
            signature = self.items()
        return hash(tuple(signature))

    def __eq__(self, other):
        if not isinstance(other, DTO):
            return False

        self_len = len(self)
        other_len = len(other)
        if self_len != other_len:
            return False

        if self_len == 0:
            return True

        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other


class CoreObject(object):
    """
    base object for all application objects.
    """

    def __init__(self):
        super().__init__()
        self.__name = None

    def __setattr__(self, name, value):
        return self.setattr(name, value)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return '{module}.{name}'.format(module=self.__module__,
                                        name=self.__class__.__name__)

    def get_name(self):
        """
        gets the name of the object.
        if name is not available, returns its class name.

        :rtype: str
        """

        if self.__name is not None:
            return self.__name
        return self.__class__.__name__

    def _set_name(self, name):
        """
        sets new name to current object.

        :param str name: object new name.
        """

        self.__name = name

    def get_doc(self):
        """
        gets docstring of the object.

        :rtype: str
        """

        return self.__doc__

    def setattr(self, name, value):
        """
        sets the given value to specified attribute.

        :param str name: attribute name.
        :param object value: attribute value.
        """

        return super().__setattr__(name, value)


class Context(DTO):
    """
    context class for storing objects in every layer.
    it's actually a dictionary with the capability to add keys directly.
    """

    def __getattr__(self, name):
        if name in self:
            return self.get(name)

        self._raise_key_error(name)

    def __getitem__(self, item):
        if item in self:
            return self.get(item)

        self._raise_key_error(item)

    def _raise_key_error(self, key):
        """
        raises an error for given key.

        :param object key: key object that caused the error.

        :raises ContextAttributeError: context attribute error.
        """

        raise ContextAttributeError('Property [{name}] not found.'.format(name=key))


class HookSingletonMeta(MultiSingletonMeta):
    """
    hook singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class Hook(CoreObject, metaclass=HookSingletonMeta):
    """
    base hook class.
    all application hook classes must be subclassed from this one.
    """
    pass


class ManagerSingletonMeta(MultiSingletonMeta):
    """
    manager singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class Manager(CoreObject, metaclass=ManagerSingletonMeta):
    """
    base manager class.
    all application manager classes must be subclassed from this one.
    """
    pass


class CLISingletonMeta(MultiSingletonMeta):
    """
    cli singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class CLI(CoreObject, metaclass=CLISingletonMeta):
    """
    base cli class.
    all application cli classes must be subclassed from this one.
    """
    pass



# -*- coding: utf-8 -*-
"""
database migration alembic exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class DatabaseMigrationAlembicManagerException(CoreException):
    """
    database migration alembic manager exception.
    """
    pass


class DatabaseMigrationAlembicManagerBusinessException(CoreBusinessException,
                                                       DatabaseMigrationAlembicManagerException):
    """
    database migration alembic manager business exception.
    """
    pass



# -*- coding: utf-8 -*-
"""
deserializer base module.
"""

from pyrin.converters.deserializer.exceptions import InvalidDeserializerTypeError
from pyrin.converters.deserializer.interface import AbstractDeserializerBase
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.globals import NULL
from pyrin.utils.string import remove_line_break_escapes


class DeserializerBase(AbstractDeserializerBase):
    """
    base deserializer class.
    """

    def __init__(self, **options):
        """
        initializes an instance of DeserializerBase.
        """

        super().__init__()
        self._next_handler = None

    def deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param object value: value to be deserialized.

        :returns: deserialized value.
        """

        deserialized_value = self._deserialize(value, **options)

        if deserialized_value is NULL:
            if self._next_handler is not None:
                return self._next_handler.deserialize(value, **options)

        if isinstance(deserialized_value, str):
            deserialized_value = remove_line_break_escapes(deserialized_value)

        return deserialized_value

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.
        this method is intended to be overridden in subclasses.

        :param object value: value to be deserialized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: deserialized value.
        """

        raise CoreNotImplementedError()

    def set_next(self, deserializer):
        """
        sets the next deserializer handler and returns it.

        :param Union[DeserializerBase, None] deserializer: deserializer instance to
                                                           be set as next handler.

        :rtype: DeserializerBase
        """

        if deserializer is not None and not isinstance(deserializer, DeserializerBase):
            raise InvalidDeserializerTypeError('Input parameter [{instance}] is not '
                                               'an instance of [{base}].'
                                               .format(instance=deserializer,
                                                       base=DeserializerBase))

        self._next_handler = deserializer
        return deserializer

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.

        :param object value: value to be deserialized.

        :rtype: bool
        """

        return isinstance(value, self.get_accepted_type())


class StringDeserializerBase(DeserializerBase):
    """
    base string deserializer class.
    """

    # these values will be used for accepted
    # formats that have no length restriction.
    UNDEF_LENGTH = None
    DEFAULT_MIN = 1
    DEFAULT_MAX = 1000000

    def __init__(self, **options):
        """
        initializes an instance of StringDeserializerBase.

        :keyword list[tuple(str, int)] accepted_formats: custom formats and their length
                                                         that this deserializer can
                                                         deserialize value from.

        :type accepted_formats: list[tuple(str format, int length)]
        """

        super().__init__(**options)

        self._accepted_formats = self.get_default_formats()

        custom_accepted_formats = options.get('accepted_formats', [])
        self._accepted_formats.extend(custom_accepted_formats)

        # min and max accepted length of strings
        # to be deserialized by this deserializer.
        self._min_length, self._max_length = self._calculate_accepted_length()

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.

        :param object value: value to be deserialized.

        :rtype: bool
        """

        if DeserializerBase.is_deserializable(self, value, **options) \
                and self.is_valid_length(value):
            return True

        return False

    def get_accepted_type(self):
        """
        gets the accepted type for this deserializer
        which could deserialize values from this type.

        :rtype: type
        """

        return str

    def get_accepted_length(self):
        """
        gets the min and max accepted length of strings to be
        deserialized by this deserializer.

        :returns tuple(int min, int max)

        :rtype: tuple(int, int)
        """

        return self._min_length, self._max_length

    def get_accepted_formats(self):
        """
        gets the accepted string formats that this deserializer
        can deserialize value from.

        :returns: list(tuple(str format, int length))

        :rtype: list(tuple(str, int))
        """

        return self._accepted_formats

    def is_valid_length(self, value):
        """
        gets a value indicating that input value has valid
        length to be deserialized by this deserializer.

        :param str value: value to be deserialized.

        :rtype: bool
        """

        length = len(value.strip())
        min_length, max_length = self.get_accepted_length()

        if length < min_length or length > max_length:
            return False

        return True

    def _calculate_accepted_length(self):
        """
        calculates the min and max accepted length of values
        to be deserialized by this deserializer.

        :returns: tuple(int min, int max)

        :rtype: tuple(int, int)
        """

        # if there is any format with length=UNDEF_LENGTH,
        # we should not enforce length restriction on values.
        if self.UNDEF_LENGTH in [item[1] for item in self.get_accepted_formats()]:
            return self.DEFAULT_MIN, self.DEFAULT_MAX

        return min([item[1] for item in self.get_accepted_formats()]), \
            max([item[1] for item in self.get_accepted_formats()])

    def get_default_formats(self):
        """
        gets default accepted formats that this
        deserializer could deserialize value from.

        :raises CoreNotImplementedError: core not implemented error.

        :return: list(tuple(str format, int length))

        :rtype: list(tuple(str, int))
        """

        raise CoreNotImplementedError()


class StringPatternDeserializerBase(StringDeserializerBase):
    """
    base string pattern deserializer class.
    this class uses regex to determine whether a value is deserializable or not.
    """

    def __init__(self, **options):
        """
        initializes an instance of StringPatternDeserializerBase.

        :keyword list[tuple(Pattern, int)] accepted_formats: custom patterns and their length
                                                             that this deserializer can
                                                             deserialize value from.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.
        if value is deserializable, the matching Pattern would be also returned.
        otherwise None would be returned instead of Pattern.

        :param object value: value to be deserialized.

        :rtype: tuple(bool, Union[Pattern, None])
        """

        if StringDeserializerBase.is_deserializable(self, value, **options):
            for pattern, length in self.get_accepted_formats():
                if pattern.match(value.strip()):
                    return True, pattern

        return False, None



# -*- coding: utf-8 -*-
"""
database services module.
"""

from pyrin.application.decorators import after_request_handler, teardown_request_handler
from pyrin.application.services import get_component
from pyrin.database import DatabasePackage


def get_current_store():
    """
    gets current database store.

    :returns: database session
    :rtype: Session
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_current_store()


def get_session_factory(request_bounded=None):
    """
    gets database session factory based on given input.
    this method should not be used directly for data manipulation.
    use `get_current_store` method instead.

    :param bool request_bounded: a value indicating that the session
                                 factory should be bounded into request.
                                 if not provided, it gets the current
                                 valid session factory.

    :returns: database session factory
    :rtype: Session
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_session_factory(request_bounded)


@after_request_handler()
def finalize_transaction(response):
    """
    this method will finalize database transaction of each request.
    we should not raise any exception in request handlers, so we return
    an error response in case of any exception.
    note that normally you should never call this method manually.

    :param CoreResponse response: response object.

    :rtype: CoreResponse
    """

    return get_component(DatabasePackage.COMPONENT_NAME).finalize_transaction(response)


@teardown_request_handler()
def cleanup_session(exception):
    """
    this method will cleanup database session of each request in
    case of any unhandled exception. we should not raise any exception
    in teardown request handlers, so we just log the exception.
    note that normally you should never call this method manually.

    :param Exception exception: exception instance.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).cleanup_session(exception)


def register_session_factory(instance, **options):
    """
    registers a new session factory or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's is_request_bounded() is already available
    in registered session factories.

    :param AbstractSessionFactoryBase instance: session factory to be registered.
                                                it must be an instance of
                                                AbstractSessionFactoryBase.

    :keyword bool replace: specifies that if there is another registered
                           session factory with the same is_request_bounded(),
                           replace it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidSessionFactoryTypeError: invalid session factory type error.
    :raises DuplicatedSessionFactoryError: duplicated session factory error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_session_factory(instance,
                                                                                  **options)


def register_bind(entity, bind_name, **options):
    """
    binds the given model class with specified bind database.

    :param CoreEntity entity: CoreEntity subclass to be bounded.
    :param str bind_name: bind name to be associated with the model class.

    :raises InvalidEntityTypeError: invalid entity type error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_bind(entity, bind_name,
                                                                       **options)


def configure_session_factories():
    """
    configures all application session factories.
    normally, you should not call this method manually.

    :raises InvalidDatabaseBindError: invalid database bind error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).configure_session_factories()


def get_default_engine():
    """
    gets database default engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_default_engine()


def get_bounded_engines():
    """
    gets database bounded engines.

    :returns: dict(str bind_name: Engine engine)
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bounded_engines()


def get_entity_to_engine_map():
    """
    gets entity to engine map.

    :returns: dict(type entity, Engine engine)
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_entity_to_engine_map()


def get_table_name_to_engine_map():
    """
    gets table name to engine map.

    :returns: dict(str table_name, Engine engine)
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_table_name_to_engine_map()


def register_hook(instance):
    """
    registers the given instance into database hooks.

    :param DatabaseHookBase instance: database hook instance to be registered.

    :raises InvalidDatabaseHookTypeError: invalid database hook type error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_hook(instance)


def get_table_engine(table_name):
    """
    gets the database engine which the provided table name is bounded to.

    :param str table_name: table name to get its bounded engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_table_engine(table_name)


def get_bind_name_engine(bind_name):
    """
    gets the database engine which the provided bind name is bounded to.

    :param str bind_name: bind name to get its bounded engine.

    :raises InvalidDatabaseBindError: invalid database bind error.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bind_name_engine(bind_name)


def get_entity_engine(entity):
    """
    gets the database engine which the provided entity class is bounded to.

    :param CoreEntity entity: entity class to get its bounded engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_entity_engine(entity)


def get_bind_config_section_name(bind_name):
    """
    gets the bind config section name for given bind name and currently
    active environment in 'database.config' from 'database.binds.config' file.

    :param str bind_name: bind name to get its section name.

    :rtype: str
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bind_config_section_name(bind_name)


def get_default_database_name():
    """
    gets default database name.

    :rtype: str
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_default_database_name()



# -*- coding: utf-8 -*-
"""
application container module.
"""

from threading import Lock

from pyrin.application.exceptions import ApplicationInstanceAlreadySetError


# holds the instance of running application.
__app__ = None
__lock__ = Lock()


def _set_app(app):
    """
    sets the global instance of running application.
    this instance should not be accessed directly from packages outside
    of `pyrin.application` package. but some internal `pyrin` packages could use
    this instance if necessary for overcoming circular dependency problem.
    outside packages, instead should always access application
    functionality through `pyrin.application.services` module.

    :raises ApplicationInstanceAlreadySetError: application instance already set error.

    :param Application app: application instance.
    """

    global __app__, __lock__

    try:
        __lock__.acquire()
        if __app__ is not None:
            raise ApplicationInstanceAlreadySetError('Application instance has been already '
                                                     'set, it could not be overwritten.')
        __app__ = app

    finally:
        if __lock__.locked():
            __lock__.release()


def _get_app():
    """
    gets the current instance of running application.
    this instance should not be accessed directly from packages outside
    of `pyrin.application` package. but some internal `pyrin` packages could use
    this instance if necessary for overcoming circular dependency problem.
    outside packages, instead should always access application
    functionality through `pyrin.application.services` module.

    :rtype: Application
    """

    global __app__
    return __app__



# -*- coding: utf-8 -*-
"""
database migration hooks module.
"""

import pyrin.database.migration.services as migration_services
import pyrin.configuration.services as config_services
import pyrin.application.services as application_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase
from pyrin.database.decorators import database_hook
from pyrin.database.hooks import DatabaseHookBase
from pyrin.utils.custom_print import print_warning, print_info


@database_hook()
class DatabaseHook(DatabaseHookBase):
    """
    database hook class.
    """

    def __init__(self):
        """
        initializes an instance of DatabaseHook.
        """

        super().__init__()

    def after_session_factories_configured(self):
        """
        this method will be called after all database
        session factories have been configured.
        """

        migration_services.configure_migration_data()


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def __init__(self):
        """
        initializes an instance of ApplicationHook.
        """

        super().__init__()

    def after_application_loaded(self):
        """
        this method will be called after application has been loaded.
        """

        if application_services.is_scripting_mode() is False:
            if config_services.get_active('database', 'drop_on_startup') is True:
                environment = config_services.get_active('environment', 'env')
                debug = config_services.get_active('environment', 'debug')
                unit_testing = config_services.get_active('environment', 'unit_testing')

                if (environment == 'development' and debug is True) or \
                        (environment == 'testing' and unit_testing is True):
                    print_warning('Dropping all models...')
                    migration_services.drop_all()

            if config_services.get_active('database', 'create_on_startup') is True:
                print_info('Creating all models...')
                migration_services.create_all()



# -*- coding: utf-8 -*-
"""
security utils secure_random module.
"""

import secrets

import pyrin.configuration.services as config_services


def get_bytes(**options):
    """
    gets a secure random bytes with given length.
    the result value should not be decoded to string, because
    it's not a safe-string and may cause an error.
    if you want string representation, use `get_hex` or `get_url_safe` methods.

    :keyword int length: length of random bytes to be get.
                         if not provided, `default_secure_random_size` config will be used.

    :rtype: bytes
    """

    length = options.get('length', None)
    if length is None:
        length = config_services.get('security', 'general',
                                     'default_secure_random_size')

    return secrets.token_bytes(length)


def get_hex(**options):
    """
    gets a secure random hex string with given length.

    :keyword int length: length of random string to be get in bytes.
                         if not provided, `default_secure_random_size` config will be used.

    :rtype: str
    """

    length = options.get('length', None)
    if length is None:
        length = config_services.get('security', 'general',
                                     'default_secure_random_size')

    return secrets.token_hex(length)


def get_url_safe(**options):
    """
    gets a secure random url-safe string with given length.

    :keyword int length: length of random string to be get in bytes.
                         if not provided, `default_secure_random_size` config will be used.

    :rtype: str
    """

    length = options.get('length', None)
    if length is None:
        length = config_services.get('security', 'general',
                                     'default_secure_random_size')

    return secrets.token_urlsafe(length)



# -*- coding: utf-8 -*-
"""
pyrin setup module.
"""

import io
import re

from setuptools import find_namespace_packages
from setuptools import setup

with io.open('README.md', 'rt', encoding='utf8') as readme_file:
    readme = readme_file.read()

with io.open('src/pyrin/__init__.py', 'rt', encoding='utf8') as version_file:
    version = re.search(r"__version__ = '(.*?)'", version_file.read()).group(1)

setup(
    name='pyrin',
    version=version,
    url='https://github.com/mononobi/pyrin',
    project_urls={
        # 'Documentation': '',
        'Code': 'https://github.com/mononobi/pyrin',
        'Issue tracker': 'https://github.com/mononobi/pyrin/issues',
    },
    license='GPL-3.0',
    author='Mohamad Nobakht',
    author_email='mohamadnobakht@gmail.com',
    maintainer='Mohamad Nobakht',
    maintainer_email='mohamadnobakht@gmail.com',
    description='Application framework for developing small to '
                'large scale enterprise applications.',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',

    ],
    packages=find_namespace_packages('src', exclude=('tests', 'tests.*')),
    package_dir={'': 'src'},
    package_data={'': ['*.config']},
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'aniso8601>=8.0.0',
        'bcrypt>=3.1.7',
        'cffi>=1.13.2',
        'pytz>=2019.3',
        'Flask>=1.1.1',
        'PyJWT>=1.7.1',
        'SQLAlchemy>=1.3.13',
        'colorama>=0.4.3',
        'python-dotenv>=0.10.5',
        'cryptography>=2.8',
        'flask-babel>=1.0.0',
        'babel>=2.8.0',
        'alembic>=1.4.0',
        'fire>=0.2.1',
        'sqlparse>=0.3.0',
    ],
    extras_require={
        'tests': [
            'pytest',
            'pytest-cov',
            'pygments',
        ],
        'docs': [
            'sphinx',
            'sphinxcontrib-log-cabinet',
            'sphinx-issues',
        ],
    },
    # entry_points={'console_scripts': ['pyrin = pyrin.cli.manager:main']},
)



# -*- coding: utf-8 -*-
"""
deserializer boolean module.
"""

import re

from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.converters.deserializer.decorators import deserializer
from pyrin.core.globals import NULL


@deserializer()
class BooleanDeserializer(StringPatternDeserializerBase):
    """
    boolean deserializer class.
    """

    # matches the bool inside string.
    # example: true, false
    # matching are case-insensitive.
    TRUE_REGEX = re.compile(r'^true$', re.IGNORECASE)
    FALSE_REGEX = re.compile(r'^false$', re.IGNORECASE)

    def __init__(self, **options):
        """
        creates an instance of BooleanDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted formats
                                                             and their length for boolean
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

        self._converter_map = self._get_converter_map()

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: bool
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        return self._converter_map[pattern]

    def get_default_formats(self):
        """
        gets default accepted formats that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(self.TRUE_REGEX, 4),
                (self.FALSE_REGEX, 5)]

    def _get_converter_map(self):
        """
        gets converter map dictionary.

        :returns: dict(Pattern format: bool value)

        :rtype: dict
        """

        return {self.TRUE_REGEX: True,
                self.FALSE_REGEX: False}



# -*- coding: utf-8 -*-
"""
locale babel manager module.
"""

from pyrin.cli.mixin import CLIMixin
from pyrin.core.context import Manager
from pyrin.globalization.locale.babel.interface import BabelCLIHandlerBase


class LocaleBabelManager(Manager, CLIMixin):
    """
    locale babel manager class.
    """

    _cli_handler_type = BabelCLIHandlerBase



# -*- coding: utf-8 -*-
"""
database migration alembic interface module.
"""

import pyrin.configuration.services as config_services

from pyrin.cli.interface import CLIHandlerBase, CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic import DatabaseMigrationAlembicPackage


class AlembicCLIHandlerBase(CLIHandlerBase):
    """
    alembic cli handler base class.
    all alembic cli handlers must be subclassed from this.
    """

    def __init__(self, name):
        """
        initializes an instance of AlembicCLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.
        """

        super().__init__(name)
        self._config_file_path = config_services.get_file_path(
            DatabaseMigrationAlembicPackage.ALEMBIC_CONFIG_STORE)

    def _inject_common_cli_options(self, commands):
        """
        injecting some common cli options into the given list.

        :param list commands: a list of all commands and their
                              values to be sent to cli command.
        """

        bounded_options = ['alembic', '-c', self._config_file_path]
        for i in range(len(bounded_options)):
            commands.insert(i, bounded_options[i])

    def _generate_common_cli_handler_options_metadata(self):
        """
        generates common cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        help_option = CLIHandlerOptionsMetadata('help', None, {True: '--help', False: None})

        options = [help_option]

        return options



# -*- coding: utf-8 -*-
"""
caching handlers simple module.
"""

import pyrin.configuration.services as config_services

from pyrin.caching.handlers.base import DictCachingHandlerBase


class SimpleCachingHandler(DictCachingHandlerBase):
    """
    simple caching handler class.
    """

    def __init__(self, **options):
        """
        initializes an instance of DictCachingHandlerBase.
        """

        options.update(timeout=config_services.get('caching', 'simple', 'timeout'),
                       max_length=config_services.get('caching', 'simple', 'max_length'))

        super().__init__('simple', **options)



# -*- coding: utf-8 -*-
"""
database migration alembic cli module.
"""

import pyrin.globalization.datetime.services as datetime_service

from pyrin.cli.decorators import cli
from pyrin.core.context import CLI


class AlembicCLI(CLI):
    """
    alembic cli class.
    this class exposes all alembic cli commands.
    """

    @cli
    def branches(self, verbose=False, help=False):
        """
        show current branch points.

        :param bool verbose: use more verbose output.
        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def current(self, verbose=False, help=False):
        """
        display the current revision for a database.

        :param bool verbose: use more verbose output.
        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def downgrade(self, revision=None, sql=False, tag=None, help=False):
        """
        revert to a previous version.
        use `base` to revert all revisions.

        :param str revision: revision identifier.

        :param bool sql: don't emit sql to database, dump to standard
                         output/file instead. see docs on offline mode.

        :param str tag: arbitrary `tag` name. can be used
                        by custom `env.py` scripts.

        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def heads(self, resolve_dependencies=False, verbose=False, help=False):
        """
        show current available heads in the script directory.

        :param bool resolve_dependencies: treat dependency versions as down revisions.
        :param bool verbose: use more verbose output.
        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def history(self, revision_range=None, indicate_current=False,
                verbose=False, help=False):
        """
        list changeset scripts in chronological order.

        :param str revision_range: specify a revision range.
                                   format is [start]:[end]

        :param bool indicate_current: indicate the current revision.
        :param bool verbose: use more verbose output.
        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def merge(self, revisions=None, message=None,
              branch_label=None, revision_id=None, help=False):
        """
        merge two revisions together. creates a new migration file.

        :param Union[str, list[str]] revisions: one or more revisions,
                                                or `heads` for all heads.

        :param str message: message string to use with `revision`.

        :param str branch_label: specify a branch label to
                                 apply to the new revision.

        :param str revision_id: specify a hardcoded revision
                                id instead of generating one.

        :param bool help: show the help message for this command.
        """

        if message is None:
            message = datetime_service.get_current_timestamp(date_sep=None,
                                                             main_sep=None,
                                                             time_sep=None)

        return dict(message=message)

    @cli
    def revision(self, message=None, autogenerate=True, sql=False,
                 head=None, splice=False, branch_label=None, version_path=None,
                 revision_id=None, depends_on=None, help=False):
        """
        create a new revision file.

        :param str message: message string to use with `revision`.

        :param bool autogenerate: populate revision script with candidate migration
                                  operations, based on comparison of database to model.

        :param bool sql: don't emit sql to database, dump to standard
                         output/file instead. see docs on offline mode.

        :param str head: specify head revision or <branchname>@head
                         to base new revision on.

        :param bool splice: allow a non-head revision as the `head` to splice onto.

        :param str branch_label: specify a branch label to
                                 apply to the new revision.

        :param str version_path: specify specific path from config for version file.
        :param str revision_id: specify a hardcoded revision id instead of generating one.

        :param Union[str, list[str]] depends_on: specify one or more revision identifiers
                                                 which this revision should depend on.

        :param bool help: show the help message for this command.
        """

        if message is None:
            message = datetime_service.get_current_timestamp(date_sep=None,
                                                             main_sep=None,
                                                             time_sep=None)

        return dict(message=message)

    @cli
    def show(self, revision=None, help=False):
        """
        show the revision(s) denoted by the given symbol.

        :param str revision: revision identifier.
                             it could be a part of revision identifier
                             to filter all matching revisions.

        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def stamp(self, revisions=None, sql=False, tag=None, purge=False, help=False):
        """
        stamp the revision table with the given revision(s).
        don't run any migrations.

        :param Union[str, list[str]] revisions: one or more revisions,
                                                or `heads` for all heads.

        :param bool sql: don't emit sql to database, dump to standard
                         output/file instead. see docs on offline mode.

        :param str tag: arbitrary `tag` name. can be used
                        by custom `env.py` scripts.

        :param bool purge: unconditionally erase the version table before stamping.
        :param bool help: show the help message for this command.
        """
        pass

    @cli
    def upgrade(self, revision=None, sql=False, tag=None, help=False):
        """
        upgrade to a later version.
        use `head` to upgrade to latest revision.

        :param str revision: revision identifier.

        :param bool sql: don't emit sql to database, dump to standard
                         output/file instead. see docs on offline mode.

        :param str tag: arbitrary `tag` name. can be used
                        by custom `env.py` scripts.

        :param bool help: show the help message for this command.
        """
        pass



# -*- coding: utf-8 -*-
"""
database migration alembic handlers base module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


class AlembicReportingCLIHandlerBase(AlembicCLIHandlerBase):
    """
    alembic reporting cli handler base class.
    all alembic reporting cli handlers must be subclassed from this.
    """

    def _generate_common_cli_handler_options_metadata(self):
        """
        generates common cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        common_options = super()._generate_common_cli_handler_options_metadata()
        verbose = CLIHandlerOptionsMetadata('verbose', None, {True: '--verbose', False: None})
        common_options.extend([verbose])

        return common_options


class AlembicUpgradeDowngradeCLIHandlerBase(AlembicCLIHandlerBase):
    """
    alembic upgrade downgrade cli handler base class.
    all alembic upgrade or downgrade cli handlers must be subclassed from this.
    """

    def _generate_common_cli_handler_options_metadata(self):
        """
        generates common cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        common_options = super()._generate_common_cli_handler_options_metadata()
        sql = CLIHandlerOptionsMetadata('sql', None, {True: '--sql', False: None})
        tag = CLIHandlerOptionsMetadata('tag', '--tag')
        revision = CLIHandlerOptionsMetadata('revision', None)
        common_options.extend([sql, tag, revision])

        return common_options



# -*- coding: utf-8 -*-
"""
permission base module.
"""

import pyrin.security.permission.services as permission_services

from pyrin.core.context import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError


class PermissionBase(CoreObject):
    """
    permission base class.
    all application permissions must be subclassed from this.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of PermissionBase.
        """

        super().__init__()
        permission_services.register_permission(self, **options)

    def __hash__(self):
        """
        this method must be implemented in all subclasses
        to get the correct hash of current permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    def __eq__(self, other):
        """
        this method must be implemented in all subclasses to get the correct
        comparison between current and other permission for equality.

        :param PermissionBase other: other permission instance to be
                                     compared to the current one.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    def __ne__(self, other):
        """
        this method must be implemented in all subclasses to get the correct
        comparison between current and other permission for not equality.

        :param PermissionBase other: other permission instance to be
                                     compared to the current one.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    def __str__(self):
        """
        this method must be implemented in all subclasses to
        get the correct string representation of current permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def __repr__(self):
        """
        this method must be implemented in all subclasses to
        get the correct string representation of current permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def to_entity(self):
        """
        gets the equivalent entity of current permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: CoreEntity
        """

        raise CoreNotImplementedError()

    def get_id(self):
        """
        gets permission id.
        note that this object must be fully unique for each different permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
hashing package.
"""

from pyrin.packaging.context import Package


class HashingPackage(Package):
    """
    hashing package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.hashing.component'



# -*- coding: utf-8 -*-
"""
database migration alembic handlers package.
"""

from pyrin.packaging.context import Package


class DatabaseMigrationAlembicHandlersPackage(Package):
    """
    database migration alembic handlers package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
permission component module.
"""

from pyrin.application.decorators import component
from pyrin.security.permission import PermissionPackage
from pyrin.security.permission.manager import PermissionManager
from pyrin.application.context import Component


@component(PermissionPackage.COMPONENT_NAME)
class PermissionComponent(Component, PermissionManager):
    """
    permission component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli interface module.
"""

import subprocess

from subprocess import CalledProcessError
from threading import Lock

from pyrin.core.context import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.globals import LIST_TYPES
from pyrin.utils.singleton import MultiSingletonMeta
from pyrin.cli.exceptions import MetaDataOptionsParamNameIsRequiredError, \
    ParamValueIsNotMappedToCLIError, InvalidCLIHandlerNameError, InvalidOptionsMetaDataTypeError


class CLIHandlerSingletonMeta(MultiSingletonMeta):
    """
    cli handler singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractCLIHandlerBase(CoreObject, metaclass=CLIHandlerSingletonMeta):
    """
    abstract cli handler base class.
    """

    def get_name(self):
        """
        gets this handlers name, the handler will be registered with this name.
        the name must be the exact command name that this handler handles.
        for example `revision`.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def execute(self, **options):
        """
        executes the current cli command with given inputs.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: execution result.
        """

        raise CoreNotImplementedError()


class CLIHandlerOptionsMetadata(CoreObject):
    """
    cli handler options metadata class.
    """

    def __init__(self, param_name,
                 cli_option_name=None,
                 param_value_to_cli_map=None,
                 default=None):
        """
        initializes an instance of CLIHandlerOptionsMetadata.

        :param str param_name: param name presented in method signature.

        :param Union[str, None] cli_option_name: relevant cli option name to `param_name`.
                                                 it could be None if command does
                                                 not have a name for this argument.
                                                 for example the `--autogenerate` flag
                                                 of alembic, could be present to imply to
                                                 `True` and absent to imply to `False`.
                                                 but `--message` flag of alembic must be
                                                 present with the message value itself.

        :param Union[dict, None] param_value_to_cli_map: a dictionary containing a mapping
                                                         between different method param values
                                                         and their representation on cli. for
                                                         example the `--autogenerate` flag of
                                                         alembic, could be present to imply to
                                                         `True` and absent to imply to `False`.
                                                         so this dict should contain two keys,
                                                         one is `True` and its value must be
                                                         '--autogenerate' and another is `False`
                                                         with the `None` value to prevent
                                                         emitting to cli.
                                                         but `--message` flag of alembic does
                                                         not have a mapping, so its dict must
                                                         be empty and instead the
                                                         `cli_option_name` must be set to
                                                         `--message` and then the exact value
                                                         of method parameter will be emitted
                                                         as the value of this flag.

        :param Union[str, None] default: default value to be emitted to cli if this
                                         param is not available. if set to None, this
                                         param will not be emitted at all.

        :raises MetaDataOptionsParamNameIsRequiredError: cli handler metadata
                                                         options param name
                                                         is required error.
        """

        super().__init__()

        if param_name in (None, '') or param_name.isspace():
            raise MetaDataOptionsParamNameIsRequiredError('CLI handler options '
                                                          'metadata "param_name" '
                                                          'must be provided.')

        self.param_name = param_name
        self.cli_option_name = cli_option_name
        self.default = default

        if param_value_to_cli_map is not None and len(param_value_to_cli_map) > 0:
            self.param_value_to_cli_map = param_value_to_cli_map
        else:
            self.param_value_to_cli_map = None


class CLIHandlerBase(AbstractCLIHandlerBase):
    """
    cli handler base class.
    all application cli handlers must be subclassed from this.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of CLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.
        """

        super().__init__()

        if name in (None, '') or name.isspace():
            raise InvalidCLIHandlerNameError('CLI handler name must be provided.')

        self._name = name
        self._options_meta_data = []
        self.process_options()

    def get_name(self):
        """
        gets this handlers name, the handler will be registered with this name.
        the name must be the exact command name that this handler handles.
        for example `revision`.

        :rtype: str
        """

        return self._name

    def process_options(self):
        """
        processes the options that are related to this handler.
        this method is intended to be overridden by base handlers of each category.
        every class that overrides this, must call `super().process_options()`
        after its work has been done.
        """

        self._process_options()

    def _process_options(self):
        """
        processes the options that are related to this handler.
        this method is intended to be overridden by each concrete handler.
        every class that overrides this, must call `super()._process_options()`
        after its work has been done.
        """
        pass

    def _add_options_metadata(self, item):
        """
        adds the given item into this handlers options metadata.

        :param CLIHandlerOptionsMetadata item: options metadata to be added.

        :raises InvalidOptionsMetaDataTypeError: invalid options metadata type error.
        """

        if not isinstance(item, CLIHandlerOptionsMetadata):
            raise InvalidOptionsMetaDataTypeError('Input parameter [{param}] is not an '
                                                  'instance of [{instance}].'
                                                  .format(param=str(item),
                                                          instance=CLIHandlerOptionsMetadata))

        self._options_meta_data.append(item)

    def execute(self, **options):
        """
        executes the current cli command with given inputs.

        :returns: execution result.
        """

        return self._execute(**options)

    def _execute(self, **options):
        """
        executes the current cli command with given inputs.

        :returns: execution result.
        """

        bounded_options = self._bind_cli_options(**options)
        self._inject_common_cli_options(bounded_options)
        return self._execute_on_cli(bounded_options)

    def _execute_on_cli(self, commands):
        """
        executes the current cli command with given inputs.

        :param list commands: a list of all commands and their
                              values to be sent to cli command.

        :returns: execution result.
        """

        try:
            return subprocess.check_call(commands)
        except CalledProcessError:
            pass

    def _inject_common_cli_options(self, commands):
        """
        this method is intended to be used by subclasses
        to inject some common cli options into the given list.
        subclasses must add some items into the given list if required.

        :param list commands: a list of all commands and their
                              values to be sent to cli.
        """
        pass

    def _bind_cli_options(self, **options):
        """
        binds cli options of current handler with values available in options.
        options are a mapping of real python method inputs.

        :returns: list of all command in the form of
                  [str name1, str value1, ...] or [str name2, ...]

        :raises ParamValueIsNotMappedToCLIError: param value is not mapped to cli error.

        :rtype: list
        """

        commands = [self.get_name()]
        for metadata in self._options_meta_data:
            real_value = options.get(metadata.param_name, None)

            if metadata.cli_option_name not in (None, '') and not \
                    metadata.cli_option_name.isspace() and \
                    self._should_send_to_cli(real_value, metadata.default) is True:
                commands.append(metadata.cli_option_name)

            if metadata.param_value_to_cli_map is not None:
                if real_value is not None and \
                        real_value not in metadata.param_value_to_cli_map:
                    raise ParamValueIsNotMappedToCLIError('Parameter value [{value}] '
                                                          'for parameter with name [{name}] '
                                                          'is not present in cli map, if '
                                                          'you want to always pass this '
                                                          'parameters real value to cli, you '
                                                          'must not provide any key arguments '
                                                          'into "param_value_to_cli_map" '
                                                          'input of "CLIHandlerOptionsMetadata" '
                                                          'class constructor. and also if you '
                                                          'want to not pass this argument to '
                                                          'cli on specific values, you must '
                                                          'provide those values as keys of '
                                                          '"param_value_to_cli_map" input of '
                                                          '"CLIHandlerOptionsMetadata" class '
                                                          'constructor with `None` value '
                                                          'attached to those keys.'
                                                          .format(value=real_value,
                                                                  name=metadata.param_name))

                cli_value = metadata.param_value_to_cli_map.get(real_value, metadata.default)
                if cli_value is not None:
                    commands.append(cli_value)
            elif self._should_send_to_cli(real_value, metadata.default):
                if real_value is None:
                    real_value = metadata.default

                if isinstance(real_value, LIST_TYPES):
                    real_value = ' '.join([str(item) for item in real_value])
                commands.append(real_value)

        return commands

    def _should_send_to_cli(self, real_value, default_value):
        """
        gets a value indicating that this param should be sent to cli.

        :param object real_value: real value of method param.
        :param str default_value: default value of options metadata.

        :rtype: bool
        """

        return real_value is not None or default_value is not None



# -*- coding: utf-8 -*-
"""
database migration alembic component module.
"""

from pyrin.application.decorators import component
from pyrin.database.migration.alembic import DatabaseMigrationAlembicPackage
from pyrin.database.migration.alembic.manager import DatabaseMigrationAlembicManager
from pyrin.application.context import Component


@component(DatabaseMigrationAlembicPackage.COMPONENT_NAME)
class DatabaseMigrationAlembicComponent(Component, DatabaseMigrationAlembicManager):
    """
    database migration alembic component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
database orm sql extractor package.
"""

from pyrin.packaging.context import Package


class DatabaseORMSQLExtractorPackage(Package):
    """
    database orm sql extractor package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.orm.sql.extractor.component'



# -*- coding: utf-8 -*-
"""
application context module.
"""

from threading import Lock

from flask import Request, Response, jsonify

import pyrin.utils.unique_id as uuid_utils
import pyrin.globalization.datetime.services as datetime_services
import pyrin.converters.deserializer.services as deserializer_services

from pyrin.application.exceptions import ComponentAttributeError, InvalidComponentNameError
from pyrin.core.context import Context, CoreObject, DTO
from pyrin.core.exceptions import ContextAttributeError
from pyrin.settings.static import DEFAULT_STATUS_CODE, JSONIFY_MIMETYPE, \
    APPLICATION_ENCODING, DEFAULT_COMPONENT_KEY
from pyrin.utils.singleton import UniqueSingletonMeta


class ApplicationContext(Context):
    """
    context class to hold application contextual data.
    """

    def _raise_key_error(self, key):
        """
        raises an error for given key.

        :param object key: key object that caused the error.

        :raises ContextAttributeError: context attribute error.
        """

        raise ContextAttributeError('Property [{name}] not found in application context.'
                                    .format(name=key))


class ApplicationComponent(ApplicationContext):
    """
    context class to hold application components.
    """

    def _raise_key_error(self, key):
        """
        raises an error for given key.

        :param object key: key object that caused the error.

        :raises ComponentAttributeError: component attribute error.
        """

        raise ComponentAttributeError('Component [{name}] is not available '
                                      'in application components.'.format(name=key))


class Component(CoreObject):
    """
    base component class.
    all component classes must inherit from this class and their respective manager class.
    """

    def __init__(self, component_name, **options):
        """
        initializes an instance of Component.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.
        """

        super().__init__()

        # component id is a tuple(str, object) and should be unique for each
        # instance unless it's intended to replace an already existing one.
        self._component_id = self.make_component_id(component_name, **options)

    def get_id(self):
        """
        gets the component id of this instance.

        :rtype: tuple(str, object)
        """

        return self._component_id

    @staticmethod
    def make_component_id(component_name, **options):
        """
        makes a component id based on input values and returns it.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.

        :raises InvalidComponentNameError: invalid component name.

        :rtype: tuple(str, object)
        """

        if component_name is None or len(component_name.strip()) <= 0:
            raise InvalidComponentNameError('Component name should not be None.')

        component_custom_key = options.get('component_custom_key', DEFAULT_COMPONENT_KEY)
        return component_name, component_custom_key


class CoreResponse(Response):
    """
    represents base response.
    this class should be used as server response.
    """

    # charset of the response.
    charset = APPLICATION_ENCODING

    # default status if none is provided.
    default_status = DEFAULT_STATUS_CODE

    # default mimetype if none is provided.
    default_mimetype = JSONIFY_MIMETYPE

    # function to use as response converter.
    response_converter = jsonify

    def __init__(self, response=None, **kwargs):
        super().__init__(response, **kwargs)

        self.request_id = None
        self.request_date = None
        self.response_date = datetime_services.now()
        self.user = None
        self.context = Context()

    @classmethod
    def force_type(cls, response, environ=None):
        response = cls.response_converter(response)
        return super().force_type(response, environ)

    def __str__(self):
        result = 'request id: "{request_id}", response date: "{response_date}", ' \
                 'request date: "{request_date}", user: "{user}", status_code: "{status_code}"'
        return result.format(response_date=self.response_date,
                             request_id=self.request_id,
                             request_date=self.request_date,
                             user=self.user,
                             status_code=self.status_code)

    def __hash__(self):
        return hash(self.request_id)


class CoreRequest(Request):
    """
    represents base request class.
    this class should be used for server request.
    """

    # charset of the request.
    charset = APPLICATION_ENCODING

    def __init__(self, environ, populate_request=True,
                 shallow=False, **options):
        super().__init__(environ, populate_request, shallow)

        self.request_id = uuid_utils.generate_uuid4()
        self.request_date = datetime_services.now()
        self.user = None
        self.component_custom_key = DEFAULT_COMPONENT_KEY
        self.client_ip = self._get_client_ip()
        self.safe_content_length = self._get_safe_content_length()
        self.context = Context()
        self.context.update(authorization=self._get_authorization_header())

        # holds the inputs of request. this value will be
        # calculated once per each request and cached.
        self.__inputs = None

        self._extract_locale()
        self._extract_timezone()

    def __str__(self):
        result = 'method: "{method}", route: "{route}", request id: "{request_id}", ' \
                 'request date: "{request_date}", user: "{user}", client_ip: "{client_ip}", ' \
                 'component_custom_key: "{component}"'
        return result.format(request_id=self.request_id,
                             request_date=self.request_date,
                             user=self.user,
                             client_ip=self.client_ip,
                             route=self.path,
                             method=self.method,
                             component=self.component_custom_key)

    def __hash__(self):
        return hash(self.request_id)

    def _get_client_ip(self):
        """
        gets client ip from request if available.

        :rtype: str
        """

        return self.environ.get('HTTP_X_REAL_IP', self.environ.get('REMOTE_ADDR', None))

    def _get_authorization_header(self):
        """
        gets the authorization header if available, otherwise returns None.

        :rtype: str
        """

        return self.headers.get('Authorization', None)

    def get_inputs(self, silent=False):
        """
        gets request inputs.

        not that if the same keyword is available in different param
        holders, the value will be replaced and no error will be raised.
        the replacement priority is as follows:
        query_strings -> body -> view_args -> files

        for example, if a keyword named 'sample_key' is available in both body and
        query_strings, the value of body will be available at the end.

        :param bool silent: specifies that if an error occurred on processing
                            json body, it should not raise an error.
                            defaults to False.

        :rtype: dict
        """

        if self.__inputs is None:
            converted_args = deserializer_services.deserialize(self.args)
            self.__inputs = DTO(**(converted_args or {}))
            self.__inputs.update(**(self.get_json(silent=silent) or {}))
            self.__inputs.update(**(self.view_args or {}))
            self.__inputs.update(files=self.files)

        return self.__inputs

    def _get_safe_content_length(self):
        """
        gets content bytes length of this request if available.
        otherwise returns 0.

        :rtype: int
        """

        return self.content_length or 0

    def _extract_locale(self):
        """
        extracts locale name from request query params and puts it into request context.
        """

        if 'lang' in self.args.keys():
            self.context.update(locale=self.args.get('lang'))

    def _extract_timezone(self):
        """
        extracts timezone name from request query params and puts it into request context.
        """

        if 'tz' in self.args.keys():
            self.context.update(timezone=self.args.get('tz'))


class ApplicationSingletonMeta(UniqueSingletonMeta):
    """
    application singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    instance = None
    _lock = Lock()



# -*- coding: utf-8 -*-
"""
database migration alembic manager module.
"""

from pyrin.cli.mixin import CLIMixin
from pyrin.core.context import Manager
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


class DatabaseMigrationAlembicManager(Manager, CLIMixin):
    """
    database migration alembic manager class.
    """

    _cli_handler_type = AlembicCLIHandlerBase
